import numpy as np
import os
from os.path import join
from scipy.ndimage.measurements import center_of_mass
from skimage.morphology import skeletonize
from skimage.morphology import opening, closing, disk
from skimage.feature import peak_local_max
from skimage.filters import gaussian
from skimage.measure import label, regionprops
import matplotlib.pyplot as plt
from itertools import product
from glob import glob
from skimage import io
from skimage.segmentation import mark_boundaries

from skimage.morphology import remove_small_objects

np.random.seed(1234567)


FILES = '/work1/project/crest/build/build_w/*/ff/*.png'
DSTDIR = './analyzer_results'
PROCESS = './process'

def min_max(x, nmin, nmax, axis=None):
    min = x.min(axis=axis, keepdims=True)
    max = x.max(axis=axis, keepdims=True)
    result = (x-min)/(max-min)*(nmax-nmin) + nmin
    return result

def edge_leaf(img, min_distance=5):
    """
    # #################################################################
    # Added by oh, 2018/07/24
    # 
    # Prameters:
    #   img: mask images removed edge area
    #   min_distance: Minimum number of pixels separating peaks in a region 
    #               of 2 * min_distance + 1 (i.e. peaks are separated by at 
    #               least min_distance). To find the maximum number of peaks,
    #               use min_distance=1.
    # Returns
    #   ret: Boolean array shaped like image, with peaks represented 
    #       by True values
    #   center_of_mass_: Calculate the center of mass of the values of 
    #       an array at labels.
    #   cone:the lower pixel value, the larger the distance between pixel and 
    #       center of plant
    # #################################################################
    """

    # #############################################################
    # Added by oh, 2018/07/24
    # Calculate the center of mass of the values of an array at labels.
    # #############################################################
    center_of_mass_ = center_of_mass(img)

    # ####################################################################
    # Added by oh, 2018/07/24
    # Move origin of coordinates in ys-xs to the center of plant
    # ####################################################################
    ys, xs = np.mgrid[:img.shape[0], :img.shape[1]].astype(np.float)  # TODO xs <-> ys
    ys -= center_of_mass_[0]
    xs -= center_of_mass_[1]

    # #############################################################
    # Added by oh, 2018/07/24
    # distance between each pixel and the center of plant
    # the lower pixel value, the larger the distance between pixel and center of
    # plant
    # #############################################################
    cone = np.sqrt(np.sqrt(ys * ys + xs * xs))
    img1 = cone * img.astype(np.float)
    nimg1 = min_max(img1, -1.0, 1.0)
    #io.imsave(basename + '_cone.png', nimg1)
    # ############################################################
    # Added by oh, 2018/07/24
    # Find peaks in an image as coordinate list or boolean mask.
    # output : ndarray or ndarray of bools
    # If indices = True : (row, column, …) coordinates of peaks.
    # If indices = False : Boolean array shaped like image, with peaks 
    # represented by True values.
    # ############################################################
    ret = peak_local_max(img1, min_distance=min_distance)
    return ret, center_of_mass_, cone


def analyse_intersection(img):
    """
    # ########################################################################
    # Added by oh, 2018/07/24
    # Obtain the number of plant, the number of holes wrapping plant, skeleton
    # image of plant and update plant mask image
    # Parameters:
    #   img: mask image removed edge of plant
    # Returns:
    #   nvertex: the number of plant area
    #   nholes: the number of holes wrapping plants
    #   skeletonized: skeleton image of plant
    #   canvas > 3: plant mask image
    # ########################################################################
    """

    # ####################################################################
    # Added by oh, 2018/07/24
    # Obtain skeleton of plant
    # skeletonize: http://scikit-image.org/docs/dev/auto_examples/edges/
    # plot_skeleton.html#sphx-glr-auto-examples-edges-plot-skeleton-py
    # ####################################################################
    skeletonized = skeletonize(img)
    h, w = skeletonized.shape[:2]
    canvas = np.zeros((h, w))

    # ######################################################################
    # Added by oh, 2018/07/24
    # Sum 8-connection pixel value into center pixel
    # ######################################################################
    for ix, iy in product((-1, 0, 1), (-1, 0, 1)):
        canvas[1:-1, 1:-1] += skeletonized[1 + iy: h - 1 + iy,
                              1 + ix: w - 1 + ix]
    # canvas = canvas > 3

    # canvas = canvas > 3
    # ################################################################
    # Added by oh, 2018/07/24
    # Label plant area
    # ################################################################
    vertices_labeled, nvertex = label(canvas > 3, return_num=True)

    # ###############################################################
    # Added by oh, 2018/07/24
    # Label hole area wrapping plants, including background
    # ################################################################
    holes_labeled, nholes = label(1 - (canvas > 0), return_num=True)
    nholes -= 1
    return nvertex, nholes, skeletonized, canvas > 3

def display_edge_leafs(img, img1, img_ori, basename):
    """
    # ####################################################################
    # Added by oh, 2018/07/24
    # Parameters:
    #   img: mask image of plant
    #   img1: mask images removed edge area
    #   img_ori: raw image of plant
    # Returns:
    #   fig: figure showing boundary image, skeleton image, mask image with 
    #       center of plant
    #   ret: Boolean array shaped like image, with peaks represented 
    #       by True values
    #   nvertices: the number of plant area
    #   nholes: the number of holes wrapping plants
    # ####################################################################
    """

    # ##################################################################
    # Added by oh, 2018/07/24
    # Label connected regions of an integer array.
    # ##################################################################
    labels = label(img1)
    print('number of regions: {:d}'.format(len(np.unique(labels))))
    scale = int(255/labels.max())
    io.imsave(basename + '_labels.png', labels*scale)
    for region in regionprops(labels):
        print("label = {0}, area = {1}".format(region.label, region.area))

    # #####################################################################
    # Added by oh, 2018/07/24
    #   nvertices: the number of plant area
    #   nholes: the number of holes wrapping plants
    #   skeletonized: skeleton image of plant
    #   intersection: plant mask image
    # #####################################################################
    nvertices, nholes, skeletonized, intersection = analyse_intersection(img1)

    # ##################################################################
    # Added by oh, 2018/07/24
    # Obtain the following value
    # ret: Boolean array shaped like image, with peaks represented 
    #       by True values
    # center_of_mass_: Calculate the center of mass of the values of 
    #       an array at labels.
    # cone:the lower pixel value, the larger the distance between pixel and 
    #       center of plant 
    # ##################################################################
    ret, center_of_mass_, cone = edge_leaf(basename, img1, min_distance=10)

    #print("vertices: {:d}, holes {:d}, leaf {:d}".format(nvertices, nholes, nvertices + 2 - 2 * nholes))  # TODO print
    #print("edge_leafs: {:d}".format(len(ret)))  # TODO print

    fig = plt.figure(figsize=(40, 20))
    ax = fig.add_subplot(1, 2, 1)
    ax.imshow(img, alpha=0.3)

    # #####################################################################
    # Added by oh, 2018/07/24
    # mark_boundaries: Return image with boundaries between labeled regions 
    # highlighted.Returns: An image in which the boundaries between labels 
    # are superimposed on the original image.
    # #####################################################################
    boundary = mark_boundaries(img, img1, color=(0, 0, 1))
    ax.imshow(boundary, alpha=0.3)
    bd_temp = min_max(boundary, -1.0, 1.0)
    io.imsave(basename + '_boundary.png', bd_temp)

    skelton = gaussian(skeletonized, sigma=2.0)
    ax.imshow(skelton, alpha=0.6, cmap='magma')
    sk_temp = min_max(skelton, -1.0, 1.0)
    io.imsave(basename + '_skelton.png', sk_temp)
    
    # ########################################################################
    # Added by oh, 2018/07/02
    # Comment out the following line if this script is run in machine without GUI
    # ########################################################################
    # ax.imshow(gaussian(intersection, sigma=2.0), alpha=0.6, cmap='spectral')

    
    # ###################################################################
    # Added by oh, 2018/07/24
    # Show center of plant in image
    # ###################################################################
    ax.imshow(cone, alpha=0.3)
    cone_temp = min_max(cone, -1.0, 1.0)
    io.imsave(basename + '_cone.png', cone_temp)

    ax.plot(ret[:, 1], ret[:, 0], '.', color='r', ms=10)
    ax.plot(center_of_mass_[1], center_of_mass_[0], 'x', color='b', ms=10)

    ax = fig.add_subplot(1, 2, 2)
    ax.imshow(img_ori)
    plt.show()
    plt.savefig(basename + '_save.png')

    return fig, ret, nvertices, nholes

def process_onefile(filename_ori, filename, dstpdf):
    """
    # ######################################################################
    # Added by oh, 2018/07/24
    # Parameters:
    #   filename_ori: raw image of plant
    #   filename: mask image of plant
    #   dstpdf: output file
    # Returns:
    #   results: dictionary
    #       ff: mask image of plant
    #       raw: raw image of plant
    #       vertices: the number of plant area
    #       holes: the number of holes wrapping plants
    #       leaf_count_vertex: the number of leaves
    #       leaf_count_edge: lenght of edge of plant
    # ######################################################################
    """
    img_ori = io.imread(filename_ori)
    img = io.imread(filename)

    # ######################################################################
    # Added by oh, 2018/07/24
    # Morphological closing on an image is defined as a dilation followed by 
    # an erosion. Closing can remove small dark spots (i.e. “pepper”) and 
    # connect small bright cracks.
    # disk: Generates a flat, disk-shaped structuring element.A pixel is within 
    # the neighborhood if the euclidean distance between it and the origin is 
    # no greater than radius.
    # 
    # Close plant area which are separated
    # ######################################################################
    img1 = closing(img, disk(5))

    # #####################################################################
    # Added by oh, 2018/07/24
    # Add gaussian filer into image, remove edge of plant area 
    # #####################################################################
    img1 = gaussian(img1, sigma=2.0)
    img1 = img1 > (0.50 * img1.max())
    
    # #################################################################
    # Added by oh, 2018/07/24
    # Obtain the following data
    # fig: figure showing boundary image, skeleton image, mask image with 
    #       center of plant
    # ret: Boolean array shaped like image, with peaks represented 
    #       by True values
    # nvertices: the number of plant area
    # nholes: the number of holes wrapping plants
    fig, ret, nvertices, nholes = display_edge_leafs(img, img1, img_ori)
    
    # #####################################################################
    # Added by oh, 2018/07/24
    # Save the number of plant, the number of holes and the number of leaves
    # into text string
    # #####################################################################
    output1 = "vertices: {:d}, holes {:d}, leaf {:d}".format(nvertices, nholes,
            nvertices + 2 - 2 * nholes)

    # ######################################################
    # Added by oh, 2018/07/24
    # Save lenght of edge of plant into text string
    # ######################################################
    output2 = "edge_leafs: {:d}".format(len(ret))
    fig.suptitle('{}\n{}\n{}'.format(filename, output1, output2), fontsize=28)
    plt.savefig(dstpdf)

    # ########################################################
    # Added by oh, 2018/07/24
    # Save results into dictionary
    # #########################################################
    results={'ff': filename, 'raw': filename_ori, 'vertices': nvertices, 'holes': nholes,
    'leaf_count_vertex': nvertices + 2 - 2 * nholes, 'leaf_count_edge': len(ret)}
    return results

def main():

    #files = sorted(glob(FILES)) # original code
    files = glob("./ff/*.png")
    np.random.shuffle(files)

    filenames_ori = []
    for i, filename in enumerate(files):
        filename_ori = filename.replace('ff', 'raw')
        #filename_ori = filename_ori.replace('/sample-W-raw-20171114','') # origianl
        filenames_ori.append(filename_ori)

    for i, (filename, filename_ori) in enumerate(zip(files, filenames_ori)):
        img_ori = io.imread(filename_ori)
        img = io.imread(filename)

        img1 = closing(img, disk(5))
        base = './process/' + os.path.basename(filename)
        folder, ext = os.path.splitext(base)
        io.imsave(base + '_close' + ext, img1)
        img1 = gaussian(img1, sigma=2.0)
        img1 = img1 > (0.50 * img1.max())
        io.imsave(base + '_thresh' + ext, img1*255)

        fig, ret, nvertices, nholes = display_edge_leafs(img, img1, img_ori, folder)

        output1 = "vertices: {:d}, holes {:d}, leaf {:d}".format(nvertices, nholes, nvertices + 2 - 2 * nholes)
        output2 = "edge_leafs: {:d}".format(len(ret))
        fig.suptitle('{}\n{}\n{}'.format(filename, output1, output2), fontsize=28)
        plt.savefig(join(DSTDIR, '{:d}.pdf'.format(i)))


if __name__ == '__main__':
    main()
